@@ -4,7 +4,8 @@
     use POSIX;
     use Config;
     use Math::MPFR::Prec;
-    use Math::MPFR::Random; # Needs to be loaded before Math::MPFR
+    use Math::MPFR::Random; # Needs to be loaded before Math::MPFR as
+                            # it contains relevant configuration info.
 
     use constant  GMP_RNDN              => 0;
     use constant  GMP_RNDZ              => 1;
@@ -40,6 +41,12 @@
     use constant MPFR_3_1_6_OR_LATER    => Math::MPFR::Random::_MPFR_VERSION() >  196869 ? 1 : 0;
     use constant MPFR_4_0_2_OR_LATER    => Math::MPFR::Random::_MPFR_VERSION() >= 262146 ? 1 : 0;
     use constant MPFR_PV_NV_BUG         => Math::MPFR::Random::_has_pv_nv_bug();
+
+    # https://github.com/StrawberryPerl/Perl-Dist-Strawberry/issues/226
+    # Math::MPFR::Random::_buggy() was modified in version 4.34 to
+    # accommodate use of vcpkg-built gmp & mpfr libraries on MS Windows.
+    use constant WIN32_FMT_BUG          => Math::MPFR::Random::_buggy();
+
     use constant NV_IS_DOUBLEDOUBLE     => 1 + (2 ** -200) > 1 ? 1 : 0;
 
     # Inspired by https://github.com/Perl/perl5/issues/19550, which affects only perl-5.35.10:
@@ -63,6 +70,8 @@
     '-='   => \&overload_sub_eq,
     '*='   => \&overload_mul_eq,
     '/='   => \&overload_div_eq,
+    '%'    => \&overload_fmod,
+    '%='   => \&overload_fmod_eq,
     '""'   => \&overload_string,
     '>'    => \&overload_gt,
     '>='   => \&overload_gte,
@@ -77,10 +86,10 @@
     'abs'  => \&overload_abs,
     '**'   => \&overload_pow,
     '**='  => \&overload_pow_eq,
-    '<<'   => \&overload_mul_2exp,
-    '<<='  => \&overload_mul_2exp_eq,
-    '>>'   => \&overload_div_2exp,
-    '>>='  => \&overload_div_2exp_eq,
+    '<<'   => \&overload_lshift,
+    '<<='  => \&overload_lshift_eq,
+    '>>'   => \&overload_rshift,
+    '>>='  => \&overload_rshift_eq,
     'atan2'=> \&overload_atan2,
     'cos'  => \&overload_cos,
     'sin'  => \&overload_sin,
@@ -100,25 +109,26 @@ MPFR_DBL_DIG MPFR_FLT128_DIG MPFR_LDBL_DIG
 MPFR_FLAGS_ALL MPFR_FLAGS_DIVBY0 MPFR_FLAGS_ERANGE MPFR_FLAGS_INEXACT MPFR_FLAGS_NAN
 MPFR_FLAGS_OVERFLOW MPFR_FLAGS_UNDERFLOW
 MPFR_FREE_LOCAL_CACHE MPFR_FREE_GLOBAL_CACHE
-MPFR_RNDA MPFR_RNDD MPFR_RNDF MPFR_RNDN MPFR_RNDU MPFR_RNDZ MPFR_PV_NV_BUG
+MPFR_RNDA MPFR_RNDD MPFR_RNDF MPFR_RNDN MPFR_RNDU MPFR_RNDZ MPFR_PV_NV_BUG WIN32_FMT_BUG
 MPFR_VERSION MPFR_VERSION_MAJOR MPFR_VERSION_MINOR MPFR_VERSION_PATCHLEVEL MPFR_VERSION_STRING
 RMPFR_PREC_MAX RMPFR_PREC_MIN RMPFR_VERSION_NUM
 
+log_2 log_10 sind cosd tand tangent
 Rmpfr_abs Rmpfr_acos Rmpfr_acosh Rmpfr_acospi Rmpfr_acosu
 Rmpfr_add Rmpfr_add_d Rmpfr_add_q Rmpfr_add_si Rmpfr_add_ui Rmpfr_add_z
 Rmpfr_agm Rmpfr_ai
 Rmpfr_asin Rmpfr_asinh Rmpfr_asinpi Rmpfr_asinu
 Rmpfr_atan Rmpfr_atan2 Rmpfr_atan2pi Rmpfr_atan2u Rmpfr_atanh Rmpfr_atanpi Rmpfr_atanu
 Rmpfr_beta
-Rmpfr_buildopt_decimal_p Rmpfr_buildopt_float128_p Rmpfr_buildopt_gmpinternals_p
-Rmpfr_buildopt_sharedcache_p Rmpfr_buildopt_tls_p Rmpfr_buildopt_tune_case
-Rmpfr_can_round Rmpfr_cbrt Rmpfr_ceil Rmpfr_check_range
+Rmpfr_buildopt_decimal_p Rmpfr_buildopt_float16_p Rmpfr_buildopt_float128_p
+Rmpfr_buildopt_gmpinternals_p Rmpfr_buildopt_sharedcache_p Rmpfr_buildopt_tls_p
+Rmpfr_buildopt_tune_case Rmpfr_can_round Rmpfr_cbrt Rmpfr_ceil Rmpfr_check_range
 Rmpfr_clear Rmpfr_clear_divby0 Rmpfr_clear_erangeflag Rmpfr_clear_flags Rmpfr_clear_inexflag
 Rmpfr_clear_nanflag Rmpfr_clear_overflow Rmpfr_clear_underflow Rmpfr_clears
 Rmpfr_cmp Rmpfr_cmp_IV Rmpfr_cmp_NV Rmpfr_cmp_d Rmpfr_cmp_f Rmpfr_cmp_float128 Rmpfr_cmp_ld
 Rmpfr_cmp_q Rmpfr_cmp_si Rmpfr_cmp_si_2exp Rmpfr_cmp_sj Rmpfr_cmp_ui Rmpfr_cmp_ui_2exp
 Rmpfr_cmp_uj Rmpfr_cmp_z Rmpfr_cmpabs Rmpfr_cmpabs_ui
-Rmpfr_compound_si
+Rmpfr_compound Rmpfr_compound_si
 Rmpfr_const_catalan Rmpfr_const_euler Rmpfr_const_log2 Rmpfr_const_pi
 Rmpfr_copysign
 Rmpfr_cos Rmpfr_cosh Rmpfr_cospi Rmpfr_cosu Rmpfr_cot Rmpfr_coth
@@ -139,8 +149,8 @@ Rmpfr_free_cache Rmpfr_free_cache2 Rmpfr_free_pool Rmpfr_frexp Rmpfr_gamma Rmpfr
 Rmpfr_get_DECIMAL128 Rmpfr_get_DECIMAL64 Rmpfr_get_FLOAT128 Rmpfr_get_IV Rmpfr_get_LD Rmpfr_get_NV
 Rmpfr_get_d Rmpfr_get_d1 Rmpfr_get_d_2exp Rmpfr_get_default_prec Rmpfr_get_default_rounding_mode
 Rmpfr_get_emax Rmpfr_get_emax_max Rmpfr_get_emax_min Rmpfr_get_emin Rmpfr_get_emin_max
-Rmpfr_get_emin_min Rmpfr_get_exp Rmpfr_get_f Rmpfr_get_float128 Rmpfr_get_flt Rmpfr_get_ld
-Rmpfr_get_ld_2exp Rmpfr_get_patches Rmpfr_get_prec Rmpfr_get_q Rmpfr_get_si Rmpfr_get_sj
+Rmpfr_get_emin_min Rmpfr_get_exp Rmpfr_get_f Rmpfr_get_float128 Rmpfr_get_flt Rmpfr_get_float16
+Rmpfr_get_ld Rmpfr_get_ld_2exp Rmpfr_get_patches Rmpfr_get_prec Rmpfr_get_q Rmpfr_get_si Rmpfr_get_sj
 Rmpfr_get_str Rmpfr_get_str_ndigits Rmpfr_get_str_ndigits_alt Rmpfr_get_ui Rmpfr_get_uj
 Rmpfr_get_version Rmpfr_get_z Rmpfr_get_z_2exp Rmpfr_get_z_exp
 Rmpfr_grandom Rmpfr_greater_p Rmpfr_greaterequal_p Rmpfr_hypot Rmpfr_inexflag_p Rmpfr_inf_p
@@ -169,7 +179,7 @@ Rmpfr_root Rmpfr_rootn_ui Rmpfr_round Rmpfr_round_nearest_away Rmpfr_roundeven
 Rmpfr_sec Rmpfr_sech Rmpfr_set Rmpfr_set_DECIMAL128 Rmpfr_set_DECIMAL64 Rmpfr_set_FLOAT128
 Rmpfr_set_IV Rmpfr_set_LD Rmpfr_set_NV Rmpfr_set_d Rmpfr_set_default_prec
 Rmpfr_set_default_rounding_mode Rmpfr_set_divby0 Rmpfr_set_emax Rmpfr_set_emin Rmpfr_set_erangeflag
-Rmpfr_set_exp Rmpfr_set_f Rmpfr_set_float128 Rmpfr_set_flt Rmpfr_set_inexflag Rmpfr_set_inf
+Rmpfr_set_exp Rmpfr_set_f Rmpfr_set_float128 Rmpfr_set_flt Rmpfr_set_float16 Rmpfr_set_inexflag Rmpfr_set_inf
 Rmpfr_set_ld Rmpfr_set_nan Rmpfr_set_nanflag Rmpfr_set_overflow Rmpfr_set_prec Rmpfr_set_prec_raw
 Rmpfr_set_q Rmpfr_set_si Rmpfr_set_si_2exp Rmpfr_set_sj Rmpfr_set_sj_2exp Rmpfr_set_str Rmpfr_set_ui
 Rmpfr_set_ui_2exp Rmpfr_set_uj Rmpfr_set_uj_2exp Rmpfr_set_underflow Rmpfr_set_z Rmpfr_set_z_2exp
@@ -179,20 +189,20 @@ Rmpfr_snprintf Rmpfr_sprintf Rmpfr_sqr Rmpfr_sqrt Rmpfr_sqrt_ui Rmpfr_strtofr
 Rmpfr_sub Rmpfr_sub_d Rmpfr_sub_q Rmpfr_sub_si Rmpfr_sub_ui Rmpfr_sub_z Rmpfr_subnormalize
 Rmpfr_sum Rmpfr_swap
 Rmpfr_tan Rmpfr_tanh Rmpfr_tanpi Rmpfr_tanu
-Rmpfr_total_order_p Rmpfr_trunc
+Rmpfr_total_order_p Rmpfr_trigamma Rmpfr_trunc
 Rmpfr_ui_div Rmpfr_ui_pow Rmpfr_ui_pow_ui Rmpfr_ui_sub Rmpfr_underflow_p Rmpfr_unordered_p
 Rmpfr_urandom Rmpfr_urandomb Rmpfr_y0 Rmpfr_y1 Rmpfr_yn
 Rmpfr_z_div Rmpfr_z_sub Rmpfr_zero_p Rmpfr_zeta Rmpfr_zeta_ui
 TRmpfr_inp_str TRmpfr_out_str
 anytoa atodouble atonum atonv
-check_exact_decimal decimalize doubletoa
-fr_cmp_q_rounded mpfr_max_orig_len mpfr_min_inter_prec mpfrtoa numtoa nvtoa nvtoa_test
-prec_cast q_add_fr q_cmp_fr q_div_fr q_mul_fr q_sub_fr rndna
+check_exact_decimal decimalize doubletoa dragon_test
+fr_cmp_q_rounded mpfr_max_orig_len mpfr_min_inter_prec mpfrtoa numtoa nvtoa nv2mpfr nvtoa_test
+prec_cast q_add_fr q_cmp_fr q_div_fr q_fmod_fr q_mul_fr q_sub_fr rndna
 );
 
     @Math::MPFR::EXPORT_OK = (@tags, 'bytes');
 
-    our $VERSION = '4.28';
+    our $VERSION = '4.38';
     #$VERSION = eval $VERSION;
 
     Math::MPFR->DynaLoader::bootstrap($VERSION);
@@ -214,6 +224,15 @@ $Math::MPFR::doubletoa_fallback = 0; # If FALLBACK_NOTIFY is defined, this scala
                                      # to produce correct result, and thus falls back to its designated
                                      # fallback routine. (See the doubletoa documentation for details.)
 
+$Math::MPFR::PERL_INFNAN = 0;
+
+my $pinf = 1e5000;      # +infinity
+my $ninf = -$pinf;      # -infinity
+my $nanv = $pinf/$pinf; # not a number
+$Math::MPFR::pinfstr = "$pinf"; # perl's string representation of +infinity
+$Math::MPFR::ninfstr = "$ninf"; # perl's string representation of -infinity
+$Math::MPFR::nanvstr = "$nanv"; # perl's string representation of a nan
+
 %Math::MPFR::NV_properties = _get_NV_properties();
 
 my %bytes = (53   =>  \&_d_bytes,
@@ -443,11 +462,19 @@ sub new {
 
 sub Rmpfr_printf {
     if(@_ == 3){
-      if(_itsa($_[1]) == 2) {wrap_mpfr_printf_rnd(@_)}
+      if(_itsa($_[1]) == 2) {wrap_mpfr_printf_rnd(@_)} # $_[1] is rounding argument (IOK).
       else {die "The second (of 3) arguments given to Rmpfr_printf() is not a valid rounding argument"}
     }
-    else {die "Rmpfr_printf must take 2 or 3 arguments: format string, [rounding,], and variable" if @_ != 2;
-    wrap_mpfr_printf(@_)}
+    else {
+      die "Rmpfr_printf must take 2 or 3 arguments: format string, [rounding,], and variable" if @_ != 2;
+      my $revised = _rewrite_fmt_arg($_[0], $_[1]);
+      if($revised) {
+        wrap_mpfr_printf($revised, nv2mpfr($_[1]));
+      }
+      else {
+        wrap_mpfr_printf(@_);
+      }
+    }
 }
 
 sub Rmpfr_fprintf {
@@ -455,20 +482,41 @@ sub Rmpfr_fprintf {
       if(_itsa($_[2]) == 2) {wrap_mpfr_fprintf_rnd(@_)}
       else {die "The third (of 4) arguments given to Rmpfr_fprintf() is not a valid rounding argument"}
     }
-    else {die "Rmpfr_fprintf must take 3 or 4 arguments: filehandle, format string, [rounding,], and variable" if @_ != 3;
-    wrap_mpfr_fprintf(@_)}
+    else {
+      die "Rmpfr_fprintf must take 3 or 4 arguments: filehandle, format string, [rounding,], and variable" if @_ != 3;
+      my $revised = _rewrite_fmt_arg($_[1], $_[2]);
+      if($revised) {
+        wrap_mpfr_fprintf($_[0], $revised, nv2mpfr($_[2]));
+      }
+      else {
+        wrap_mpfr_fprintf(@_);
+      }
+    }
 }
 
 sub Rmpfr_sprintf {
     my $len;
     if(@_ == 5){
-      if(_itsa($_[2]) == 2) {
+      if(_itsa($_[2]) == 2) { # IV
         $len = wrap_mpfr_sprintf_rnd(@_);
         return $len;
       }
       else {die "The third (of 5) arguments given to Rmpfr_sprintf() is not a valid rounding argument"}
     }
-    die "Rmpfr_sprintf must take 4 or 5 arguments: buffer, format string, [rounding,], variable and buffer size" if @_ != 4;
+    die "Rmpfr_sprintf must take 4 or 5 arguments: buffer, format string, [rounding,] variable and buffer size" if @_ != 4;
+
+    my $revised = _rewrite_fmt_arg($_[1], $_[2]);
+    if($revised) {
+      $len = wrap_mpfr_sprintf($_[0], $revised, nv2mpfr($_[2]), $_[3]);
+      return $len;
+    }
+
+# Alternatively, this should work, but it crashes intermittently
+#    if(!_hex_fmt_ok($_[1], $_[2])) {
+#      $len = _gmp_sprintf_nv(@_);
+#      return $len;
+#    }
+
     $len = wrap_mpfr_sprintf(@_);
     return $len;
 }
@@ -476,13 +524,20 @@ sub Rmpfr_sprintf {
 sub Rmpfr_snprintf {
     my $len;
     if(@_ == 6){
-      if(_itsa($_[3]) == 2) {
+      if(_itsa($_[3]) == 2) {  # IV
         $len = wrap_mpfr_snprintf_rnd(@_);
         return $len;
       }
       else {die "The fourth (of 6) arguments given to Rmpfr_snprintf() is not a valid rounding argument"}
     }
     die "Rmpfr_snprintf must take 5 or 6 arguments: buffer, bytes written, format string, [rounding,], variable and buffer size" if @_ != 5;
+
+    my $revised = _rewrite_fmt_arg($_[2], $_[3]);
+    if($revised) {
+      $len = wrap_mpfr_snprintf($_[0], $_[1], $revised, nv2mpfr($_[3]), $_[4]);
+      return $len;
+    }
+
     $len = wrap_mpfr_snprintf(@_);
     return $len;
 }
@@ -850,18 +905,11 @@ sub anytoa {
     unless Math::MPFR::_itsa($_[0]) == 5;
 
   my $v = shift;
-  my $bits;
+  die "anytoa() now takes only one argument" if defined($_[0]) ;
 
-  if($_[0]) {
-    $bits = shift;
-    die "2nd argument given to anytoa() must be 0 or 53 or 64 or 113 or 2098"
-    unless ($bits == 53 || $bits == 64 || $bits == 113 || $bits == 2098);
-  }
-  else {
-    $bits = Rmpfr_get_prec($v);
-    die "Precision of arg given to anytoa() must be 53 or 64 or 113 or 2098"
-    unless ($bits == 53 || $bits == 64 || $bits == 113 || $bits == 2098);
-  }
+  my $bits = Rmpfr_get_prec($v);
+  die "Precision of arg given to anytoa() must be 53 or 64 or 113 or 2098"
+  unless ($bits == 53 || $bits == 64 || $bits == 113 || $bits == 2098);
 
   my $emax = Rmpfr_get_emax();                # Save original value
   my $emin = Rmpfr_get_emin();                # Save original value
@@ -898,7 +946,7 @@ sub anytoa {
       Rmpfr_set_d($f, $msd, MPFR_RNDN);
       Rmpfr_set_emax($emax);                  # Revert to original value
       Rmpfr_set_emin($emin);                  # Revert to original value
-      return anytoa($f, 53);
+      return anytoa($f);
     }
 
     # Determine the no. of implied (intermediate)
@@ -948,9 +996,6 @@ sub anytoa {
   return mpfrtoa($f_init);
 }
 
-
-
-
 ###########################
 ###########################
 
@@ -958,6 +1003,8 @@ sub nvtoa_test {
   # 1st arg must be a string (POK);
   # 2nd arg must be either an NV (NOK) or a Math::MPFR object
 
+  warn "nvtoa_test() is DEPRECATED. Please use dragon_test() instead";
+
   die "nvtoa_test() requires at least version 3.1.6 of the MPFR library"
     if 196870 > MPFR_VERSION;
 
@@ -974,7 +1021,8 @@ sub nvtoa_test {
 
   if(ref($n) eq 'Math::MPFR') {
     $check = Rmpfr_init2(Rmpfr_get_prec($n));
-    Rmpfr_set($check, abs($n), MPFR_RNDN);
+    Rmpfr_set_str($check, $s, 10, MPFR_RNDN);
+    Rmpfr_abs($check, $check, MPFR_RNDN);
   }
   else {
     die "2nd arg to nvtoa_test() must be either an NV or a Math::MPFR object"
@@ -1111,6 +1159,177 @@ sub nvtoa_test {
   return $ret;
 }
 
+#############################
+#############################
+#############################
+#############################
+
+sub dragon_test {
+
+  # 1st arg must be either an NV (NOK) or an IV(IOK) or a Math::MPFR object
+  # If there's a 2nd arg (optional) and it matches /debug/i, then $debug
+  # is set to 1, and some debug info is output during the running of the test.
+
+  die "dragon_test() requires at least version 3.1.6 of the MPFR library"
+    if 196870 > MPFR_VERSION;
+
+  my ($repro, $op, $reco, $is_nv, $ret, $debug);
+  $debug = 1 if(@_ == 2 && $_[1] =~ /debug/i);
+
+  if(ref($_[0]) eq 'Math::MPFR') {
+    $repro = mpfrtoa($_[0]);
+  }
+  else {
+    unless (_SvNOK($_[0]) || _itsa($_[0]) < 3) { # $_[0] needs to be IV/UV/NV
+      die "1st arg to dragon_test() must be either an IV or an NV or a Math::MPFR object"
+    }
+    $is_nv = 1;
+    $repro = nvtoa($_[0]);
+  }
+
+  # Check that signs match:
+  if($repro =~ s/^\-//) {
+    die "In dragon_test(): signs do not match"
+      if $_[0] > 0;
+  }
+  else {
+    die "In dragon_test():mismatch of signs"
+      if($_[0] < 0);
+  }
+
+  # Deal only with the +ve form for simplicity.
+  if($is_nv) {
+    $op = abs($_[0]); # Can change $_[0] from NV to IV
+                      # but that doesn't matter.
+    $reco = atonv($repro);
+  }
+  else {
+    $op = Rmpfr_init2(Rmpfr_get_prec($_[0]));
+    my $inex = Rmpfr_abs($op, $_[0], MPFR_RNDN);
+    die "Copying of argument in dragon_test() failed"
+      if $inex;
+    $reco = Rmpfr_init2(Rmpfr_get_prec($_[0]));
+    Rmpfr_set_str($reco, $repro, 10, MPFR_RNDN);
+  }
+
+  print "OP: $op\nREPRO: $repro\n" if $debug;
+
+  # Handle inf, nan and zero - test that both $n and $check
+  # are the same. No further testing required.
+
+  if($op == 0) {
+    return 15 if $reco == 0;
+    return 0;
+  }
+
+  if($is_nv) {
+    if(Rmpfr_nan_p(Math::MPFR->new($op))) {
+      return 15 if $repro =~ /^nan/i;
+      return 0;
+    }
+    if(Rmpfr_inf_p(Math::MPFR->new($op))) {
+      return 15 if $repro =~ /^inf/i;
+      return 0;
+    }
+  }
+  else {
+    if(Rmpfr_nan_p($op)) {
+      return 15 if $repro =~ /^nan/i;
+      return 0;
+    }
+    if(Rmpfr_inf_p($op)) {
+      return 15 if $repro =~ /^inf/i;
+      return 0;
+    }
+  }
+
+  $ret++ if $reco == $op; # round trip successful
+
+  my @r = split /e/i, $repro;
+
+  if($debug) {
+    print "SPLIT:\n$r[0]";
+    if(defined($r[1])) { print " $r[1]\n" }
+    else { print " no exponent\n" }
+  }
+
+  # Increment $ret by 8 if and only if there are no errant trailing
+  # zeroes in $r[0] .
+
+  if(!defined($r[1])) {
+    $ret += 8 if ($r[0] =~ /\.0$/ || $r[0] !~ /0$/);
+    $r[1] = 0;       # define $r[1] by setting it to zero.
+  }
+  else {
+   $ret += 8 unless $r[0] =~ /0$/;
+  }
+
+  # We remove from $repro any trailing mantissa zeroes, and then
+  # replace the least significant digit with zero.
+  # IOW, we effectively chop off the least siginificant digit, thereby
+  # rounding it down to the next lowest decimal precision.)
+  # This altered string should assign to a value less than $op.
+
+  chop($r[0]) while $r[0] =~ /0$/;
+  $r[0] =~ s/\.$//;
+  while($r[0] =~ /0$/) {
+    chop $r[0];
+    $r[1]++;
+  }
+
+  return $ret + 6 if length($r[0]) < 2; # chop test and increment test inapplicable.
+
+  substr($r[0], -1, 1, '0');
+
+
+  my $chopped = $r[1] ? $r[0] . 'e' . $r[1]
+                      : $r[0];
+
+  print "CHOPPED:\n$chopped\n\n" if $debug;
+
+  if($is_nv) {
+    $ret += 2 if atonv($chopped) < $op; # chop test ok.
+  }
+  else {
+    Rmpfr_set_str($reco, $chopped, 10, MPFR_RNDN);
+    $ret += 2 if $reco < $op;          # chop test ok.
+  }
+
+  # Now we derive a value that is $chopped rounded up to the next lowest
+  # decimal representation.
+  # This new string should assign to a value that is greater
+  # than the given $op.
+
+  if($r[0] =~ /\./) {
+    # We must remove the '.', do the string increment,
+    # and then reinsert the '.' in the appropriate place.
+    my @mantissa = split /\./, $r[0];
+    my $point_pos = -(length($mantissa[1]));
+    my $t = $mantissa[0] . $mantissa[1];
+    $t++ for 1..10;
+    substr($t, $point_pos, 0, '.');
+    $r[0] = $t;
+  }
+  else {
+    $r[0]++ for 1..10;
+  }
+
+  my $incremented = defined($r[1]) ? $r[0] . 'e' . $r[1]
+                                   : $r[0];
+
+  print "INCREMENTED:\n$incremented\n" if $debug;
+
+  if($is_nv) {
+    $ret += 4 if atonv($incremented) > $op; # increment test ok.
+  }
+  else {
+    Rmpfr_set_str($reco, $incremented, 10, MPFR_RNDN);
+    $ret += 4 if $reco > $op;              # increment test ok.
+  }
+
+  return $ret;
+}
+
 #############################
 #############################
 
@@ -1152,7 +1371,6 @@ sub _get_exp {
 *Rmpfr_randinit_lc_2exp      = \&Math::MPFR::Random::Rmpfr_randinit_lc_2exp;
 *Rmpfr_randinit_lc_2exp_size = \&Math::MPFR::Random::Rmpfr_randinit_lc_2exp_size;
 
-
 sub nvtoa {
   # Special handling required for DoubleDouble
   # Unable to get the _nvtoa XSub to work flawlessly with DoubleDoubles,
@@ -1443,25 +1661,140 @@ sub _decrement {
 
   my $ret = $r[1] ? $r[0] . 'e' . $r[1]
                   : $r[0];
-
   return $ret;
 }
 
-#sub tz_test {
-#  # Detect any unwanted trailing zeroes
-#  # in values returned by nvtoa().
-#
-#  my $s = shift;
-#  my @r = split /e/i, $s;
-#
-#  if(!defined($r[1])) {
-#    return 1 if $r[0] =~ /\.0$/; # pass
-#    return 0 if $r[0] =~ /0$/;   # fail
-#  }
-#
-#  return 0 if $r[0] =~ /0$/;     # fail (for our formatting convention)
-#  return 1;                      # pass
-#}
+sub _rewrite_fmt_arg {
+  # Can Return a true value if and only if the constant
+  # WIN32_FMT_BUG is set to a true value, in which case it
+  # returns a rewritten format string.
+  # We allow "%a"/"%A" formatting only if nvtype is 'double' and
+  # we allow "%La"/"%LA" formatting only if nvtype is 'long double'.
+
+  my $arg = shift;
+  # First check for a match of (eg) "%La" or "%A"
+  # at the beginning of $arg.
+  if($arg =~ /^%L?[a,A]/) {
+    my $match = $&;
+    if(!Math::MPFR::_SvNOK($_[0])) { die "\"$match\" formatting applies only to NVs. Use \"%Ra\" for Math::MPFR objects." }
+    if($match =~ /L/) { die "\"$match\" formatting applies only to long doubles." unless $Config{nvtype} eq 'long double'; }
+    else { die "\"$match\" formatting applies only to doubles." unless $Config{nvtype} eq 'double'; }
+    if(WIN32_FMT_BUG) {
+      $arg =~ s/^%L?/%R/;
+      return $arg;
+    }
+    return '';
+  }
+
+  # Need to match (eg) "%La" or "%A", but also to NOT match (eg) "%%La" or "%%A".
+  if($arg =~ /[^%]%L?[a,A]/) {
+    my $match = $&;
+    if(!Math::MPFR::_SvNOK($_[0])) { die "\"", substr($match, 1), "\" formatting applies only to NVs. Use \"%Ra\" for Math::MPFR objects." }
+    if($match =~ /L/) { die "\"", substr($match, 1), "\" formatting applies only to long doubles." unless $Config{nvtype} eq 'long double'; }
+    else { die "\"", substr($match, 1), "\" formatting applies only to doubles." unless $Config{nvtype} eq 'double'; }
+    if(WIN32_FMT_BUG) {
+      my $and_init = $match;
+      my $temp = $match;
+      my $start = substr($temp, 0, 1, '');
+      $temp =~ s/^%L?/%R/;
+      $start .= $temp;
+      $arg =~ s/$and_init/$start/;
+      return $arg;
+    }
+  }
+  return ''; # Nothing to modify - no action necessary.
+}
+
+sub nv2mpfr {
+  my $prec = Rmpfr_get_default_prec();
+  Rmpfr_set_default_prec($Math::MPFR::NV_properties{bits});
+  my $ret = Math::MPFR->new($_[0]);
+  Rmpfr_set_default_prec($prec);
+  return $ret;
+}
+
+sub _win32_formatting_ok {   # Duplicated in Random/Random.pm
+    # Return 1 if either __GMP_CC or __GMP_CFLAGS
+    # include the string '-D__USE_MINGW_ANSI_STDIO'.
+    # Else return 0.
+
+    my $cc = _gmp_cc();		# __GMP_CC
+    my $cflags = _gmp_cflags();	# __GMP_CFLAGS
+
+    return 1 if ( defined($cc)     && $cc     =~/\-D__USE_MINGW_ANSI_STDIO/ );
+    return 1 if ( defined($cflags) && $cflags =~/\-D__USE_MINGW_ANSI_STDIO/ );
+    return 0;
+}
+
+sub _hex_fmt_ok {
+  # Return 1 if no hex ("%a") formatting is requested.
+  # Return 1 if the requested hex format ("%a"/"%A"/"%La"/"%LA") is appropriate.
+  # If we haven't died or returned by now, then return 0 if WIN32_FMT_BUG is set;
+  # otherwise return 1.
+  # A return of 1 tells the caller (Rmpfr_*printf) that no other action is necessary.
+  # A return of 0 tells the caller(Rmpfr_*printf) that it needs to call _rewrite_fmt_arg().
+
+  my $fmt = shift;
+
+  if($fmt =~ /^%L?[a,A]|[^%]%L?[a,A]/) {
+     $fmt = $&;
+     if(!Math::MPFR::_SvNOK($_[0])) { die "\"%a\" formatting applies only to NVs. Use \"%Ra\" for Math::MPFR objects." }
+     if($fmt =~ /L/) { die "\"%La\" formatting applies only to long doubles." unless $Config{nvtype} eq 'long double'; }
+     else { die "\"%a\" formatting applies only to doubles." unless $Config{nvtype} eq 'double'; }
+     return 0 if WIN32_FMT_BUG; # Tells the caller to call the relevant _gmp_*printf_nv() function
+     return 1;
+  }
+  else {return 1} # No %a formatting requested. Proceed as normal.
+}
+
+sub overload_lshift {
+  if($_[2] || !_looks_like_number($_[1])) {
+    die "Math::MPFR: When overloading '<<', the argument that specifies the number of bits to be shifted must be a perl number";
+  }
+  return _overload_lshift(@_) if $_[1] >= 0;
+  return _overload_rshift($_[0], -$_[1], $_[2]);
+}
+
+sub overload_lshift_eq {
+  if($_[2] || !_looks_like_number($_[1])) {
+    die "Math::MPFR: When overloading '<<=', the argument that specifies the number of bits to be shifted must be a perl number";
+  }
+  return _overload_lshift_eq(@_) if $_[1] >= 0;
+  return _overload_rshift_eq($_[0], -$_[1], $_[2]);
+}
+
+sub overload_rshift {
+  if($_[2] || !_looks_like_number($_[1])) {
+    die "Math::MPFR: When overloading '>>', the argument that specifies the number of bits to be shifted must be a perl number";
+  }
+  return _overload_rshift(@_) if $_[1] >= 0;
+  return _overload_lshift($_[0], -$_[1], $_[2]);
+}
+
+sub overload_rshift_eq {
+  if($_[2] || !_looks_like_number($_[1])) {
+    die "Math::MPFR: When overloading '>>=', the argument that specifies the number of bits to be shifted must be a perl number";
+  }
+  return _overload_rshift_eq(@_) if $_[1] >= 0;
+  return _overload_lshift_eq($_[0], -$_[1], $_[2]);
+}
+
+sub overload_fmod {
+  if(ref($_[1]) ne 'Math::MPFR') {
+    return _overload_fmod($_[0], Math::MPFR->new($_[1]), 0) unless $_[2];
+    return _overload_fmod(Math::MPFR->new($_[1]), $_[0], 0);
+  }
+  return _overload_fmod(@_);
+}
+
+sub overload_fmod_eq {
+  if(ref($_[1]) ne 'Math::MPFR') {
+    return _overload_fmod_eq($_[0], Math::MPFR->new($_[1]), 0) unless $_[2];
+    return _overload_fmod_eq(Math::MPFR->new($_[1]), $_[0], 0);
+  }
+  return _overload_fmod_eq(@_);
+}
+
 
 1;
 
